package ru.dbotthepony.mc.otm.screen;

import com.google.common.collect.Sets;
import com.mojang.blaze3d.platform.InputConstants;
import com.mojang.blaze3d.systems.RenderSystem;
import com.mojang.blaze3d.vertex.PoseStack;
import net.minecraft.ChatFormatting;
import net.minecraft.Util;
import net.minecraft.client.Minecraft;
import net.minecraft.client.gui.Font;
import net.minecraft.client.gui.components.Widget;
import net.minecraft.client.gui.screens.inventory.AbstractContainerScreen;
import net.minecraft.client.renderer.entity.ItemRenderer;
import net.minecraft.network.chat.Component;
import net.minecraft.resources.ResourceLocation;
import net.minecraft.util.Mth;
import net.minecraft.world.entity.player.Inventory;
import net.minecraft.world.inventory.AbstractContainerMenu;
import net.minecraft.world.inventory.ClickType;
import net.minecraft.world.inventory.Slot;
import net.minecraft.world.item.ItemStack;
import ru.dbotthepony.mc.otm.OverdriveThatMatters;
import ru.dbotthepony.mc.otm.menu.MatteryMenu;
import ru.dbotthepony.mc.otm.menu.PoweredMatteryMenu;
import ru.dbotthepony.mc.otm.menu.slot.BatterySlot;
import ru.dbotthepony.mc.otm.menu.slot.MatterySlot;
import ru.dbotthepony.mc.otm.menu.widget.GaugeWidget;
import ru.dbotthepony.mc.otm.screen.panels.*;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;

import static org.lwjgl.opengl.GL11.GL_ALWAYS;
import static org.lwjgl.opengl.GL11.GL_LESS;
import static org.lwjgl.opengl.GL13.GL_TEXTURE0;

/**
 * This class encapsulate most of logic for handling EditablePanel and it's children
 * It inherits AbstractContainerScreen, but most of methods of AbstractContainerScreen are re implemented or straight removed
 * since they are present inside EditablePanel subclasses, therefore it inherits
 *
 * It inherits AbstractContainerScreen just for signature compatibility
 *
 * Such as slots, they are not drawn inside MatteryScreen, but instead
 * drawn by SlotPanel(s)
 *
 * @param <T> type of menu, which extends MatteryMenu
 */
public abstract class MatteryScreen<T extends MatteryMenu> extends AbstractContainerScreen<T> {
	public interface IMatteryScreenLeftPanel {}

	public interface IMatteryScreenGaugeGetter extends IMatteryScreenLeftPanel {
		List<GaugeWidget> getGauges();
	}

	public interface IMatteryScreenBatteryGetter extends IMatteryScreenLeftPanel {
		List<MatterySlot> getBatterySlots();
	}

	public interface IMatteryScreenGridBased {
		void createGridPanels(FlexGridPanel grid);
	}

	public static final float SNAPBACK_SPEED = 100.0F;
	public static final int QUICKDROP_DELAY = 500;
	public static final int DOUBLECLICK_SPEED = 250;
	public static final int SLOT_ITEM_BLIT_OFFSET = 100;
	public static final int HOVER_ITEM_BLIT_OFFSET = 200;

	public final T menu;
	public final Component playerInventoryTitle;

	@Nullable
	public Slot hoveredSlot;
	@Nullable
	public Slot clickedSlot;
	@Nullable
	public Slot snapbackEnd;
	@Nullable
	public Slot quickdropSlot;
	@Nullable
	public Slot lastClickSlot;

	protected int leftPos;
	protected int topPos;
	protected boolean isSplittingStack;
	protected ItemStack draggingItem = ItemStack.EMPTY;
	protected int snapbackStartX;
	protected int snapbackStartY;
	protected long snapbackTime;
	protected ItemStack snapbackItem = ItemStack.EMPTY;
	protected long quickdropTime;
	public final Set<Slot> quickCraftSlots = Sets.newHashSet();

	public boolean isQuickCrafting() {
		return isQuickCrafting;
	}

	protected boolean isQuickCrafting;

	public int getQuickCraftingType() {
		return quickCraftingType;
	}

	protected int quickCraftingType;
	protected int quickCraftingButton;
	protected boolean skipNextRelease;
	protected int quickCraftingRemainder;
	protected long lastClickTime;
	protected int lastClickButton;
	protected boolean doubleclick;
	protected ItemStack lastQuickMoved = ItemStack.EMPTY;

	protected final ArrayList<EditablePanel> panels = new ArrayList<>();
	public FramePanel inventory_frame;
	public FramePanel main_frame;
	private boolean made_main_frame = false;

	public MatteryScreen(T menu, Inventory inventory, Component title) {
		super(menu, inventory, title);
		this.menu = menu;
		playerInventoryTitle = inventory.getDisplayName();

		if (menu.inventory_slots.size() != 0) {
			inventory_frame = new FramePanel(this, null, 0, 0, 18 * 9 + 16, 3 * 18 + 18 + 24, inventory.getDisplayName());
			panels.add(inventory_frame);

			for (var slot : menu.inventory_slots) {
				new SlotPanel<>(
					this,
					inventory_frame,
					slot.x,
					slot.y,
					18, 18,
					slot);
			}
		}
	}

	public void addPanel(EditablePanel panel) {
		if (!panels.contains(panel))
			panels.add(panel);
	}

	public void removePanel(EditablePanel panel) {
		panels.remove(panel);
	}

	/**
	 * Default implementation allows for builder-like creation of main frame panel
	 * achieved by this implementing specific interfaces
	 *
	 * can be freely overridden by child classes
	 *
	 * @return FramePanel created, or null
	 */
	@Nullable
	protected FramePanel makeMainFrame() {
		var frame = new FramePanel(this, null, 0, 0, 18 * 9 + 16, 100, getTitle());

		autoAttachToFrame(frame);

		return frame;
	}

	protected void autoAttachToFrame(FramePanel frame) {
		if (this instanceof IMatteryScreenLeftPanel) {
			var reserve = new EditablePanel(this, frame, 0, 0, 0, 0);
			reserve.setDock(Dock.LEFT);

			float available_height = frame.getHeight() - 14;

			if (this instanceof IMatteryScreenGaugeGetter gauges) {
				float width = 0;
				float height = 0;

				for (var gauge : gauges.getGauges()) {
					width += gauge.getImageWidth();
					height = Math.max(height, gauge.getImageHeight());
				}

				width += 2;

				var grid_gauges = new FlexGridPanel(this, reserve, 0, 0, width, height);

				if (!(this instanceof IMatteryScreenBatteryGetter))
					grid_gauges.setDock(Dock.FILL);
				else
					grid_gauges.setDock(Dock.TOP);

				for (var gauge : gauges.getGauges()) {
					new MatteryWidgetPanel(this, grid_gauges, gauge);
				}

				reserve.setWidth(Math.max(width, reserve.getWidth()));
				available_height -= height;
			}

			if (this instanceof IMatteryScreenBatteryGetter batteries) {
				var get = batteries.getBatterySlots();
				int capacity = Math.max(1, ((int) available_height) / 18);

				reserve.setWidth(Math.max((float) Math.ceil(get.size() / (float) capacity) * 18, reserve.getWidth()));

				var grid = new FlexGridPanel(this, reserve, 0, 0, width, height);
				grid.setDock(Dock.FILL);

				for (var slot : get) {
					new SlotPanel<>(this, grid, slot);
				}
			}

			reserve.setWidth(reserve.getWidth() + 4);
			reserve.setDockPadding(4, 0, 0, 0);
		}

		if (this instanceof IMatteryScreenGridBased grid_based) {
			var grid = new FlexGridPanel(this, frame, 0, 0, 0, 0);
			grid.setDock(Dock.FILL);

			grid_based.createGridPanels(grid);
		}
	}

	public ItemRenderer getItemRenderer() {
		return itemRenderer;
	}

	public Font getFont() {
		return font;
	}

	/**
	 * Moves panel at top of all other panels
	 *
	 * @param panel panel to be pushed up
	 */
	public void popup(EditablePanel panel) {
		if (!panels.contains(panel))
			throw new IllegalArgumentException("No such panel " + panel);

		if (panels.indexOf(panel) == panels.size())
			return;

		panels.remove(panel);
		panels.add(panel);
	}

	public void movePanels() {
		var top = height / 2;

		if (main_frame != null) {
			top -= main_frame.getHeight() / 2;
		}

		if (inventory_frame != null) {
			top -= inventory_frame.getHeight() / 2;
		}

		if (main_frame != null) {
			main_frame.setPos(width / 2 - main_frame.getWidth() / 2, top);
			top += main_frame.getHeight();
		}

		if (inventory_frame != null) {
			inventory_frame.setPos(width / 2 - inventory_frame.getWidth() / 2, top);
			top += inventory_frame.getHeight();
		}
	}

	@Override
	protected void init() {
		super.init();

		movePanels();
	}

	@Override
	public boolean mouseClicked(double p_97748_, double p_97749_, int p_97750_) {
		var happen_to_click = false;

		for (int i = panels.size() - 1; i >= 0; i--) {
			var panel = panels.get(i);

			if (happen_to_click) {
				panel.killFocus();
			} else if (panel.mouseClickedChecked(p_97748_, p_97749_, p_97750_)) {
				happen_to_click = true;
			}
		}

		return true;
	}

	@Override
	public boolean mouseDragged(double p_97752_, double p_97753_, int p_97754_, double p_97755_, double p_97756_) {
		for (int i = panels.size() - 1; i >= 0; i--) {
			var panel = panels.get(i);

			if (panel.mouseDraggedChecked(p_97752_, p_97753_, p_97754_, p_97755_, p_97756_)) {
				return true;
			}
		}

		return true;
	}

	@Override
	public boolean mouseReleased(double p_97812_, double p_97813_, int p_97814_) {
		for (int i = panels.size() - 1; i >= 0; i--) {
			var panel = panels.get(i);

			if (panel.mouseReleasedChecked(p_97812_, p_97813_, p_97814_)) {
				return true;
			}
		}

		// handle quick moving ("quickcraft") by ourselves if we depressed outside of slot
		onMouseReleasedSlot(null, p_97812_, p_97813_, p_97814_, true);

		return true;
	}

	@Override
	public boolean mouseScrolled(double p_94686_, double p_94687_, double p_94688_) {
		for (int i = panels.size() - 1; i >= 0; i--) {
			var panel = panels.get(i);

			if (panel.mouseScrolledChecked(p_94686_, p_94687_, p_94688_)) {
				return true;
			}
		}

		return true;
	}

	@Override
	public boolean keyReleased(int p_94715_, int p_94716_, int p_94717_) {
		for (int i = panels.size() - 1; i >= 0; i--) {
			var panel = panels.get(i);

			if (panel.keyReleased(p_94715_, p_94716_, p_94717_)) {
				return true;
			}
		}

		return true;
	}

	@Override
	public boolean charTyped(char p_94683_, int p_94684_) {
		for (int i = panels.size() - 1; i >= 0; i--) {
			var panel = panels.get(i);

			if (panel.charTyped(p_94683_, p_94684_)) {
				return true;
			}
		}

		return true;
	}

	@Override
	public boolean keyPressed(int p_97765_, int p_97766_, int p_97767_) {
		InputConstants.Key mouseKey = InputConstants.getKey(p_97765_, p_97766_);

		for (int i = panels.size() - 1; i >= 0; i--) {
			var panel = panels.get(i);

			if (panel.keyPressed(p_97765_, p_97766_, p_97767_)) {
				return true;
			}
		}

		boolean handled = this.checkHotbarKeyPressed(p_97765_, p_97766_);// Forge MC-146650: Needs to return true when the key is handled
		if (this.hoveredSlot != null && this.hoveredSlot.hasItem()) {
			if (this.minecraft.options.keyPickItem.isActiveAndMatches(mouseKey)) {
				this.slotClicked(this.hoveredSlot, this.hoveredSlot.index, 0, ClickType.CLONE);
				handled = true;
			} else if (this.minecraft.options.keyDrop.isActiveAndMatches(mouseKey)) {
				this.slotClicked(this.hoveredSlot, this.hoveredSlot.index, hasControlDown() ? 1 : 0, ClickType.THROW);
				handled = true;
			}
		} else if (this.minecraft.options.keyDrop.isActiveAndMatches(mouseKey)) {
			handled = true; // Forge MC-146650: Emulate MC bug, so we don't drop from hotbar when pressing drop without hovering over a item.
		}

		// return handled;

		// nothing dangerous in super method
		return super.keyPressed(p_97765_, p_97766_, p_97767_);
	}

	public void renderFloatingItem(ItemStack p_97783_, int p_97784_, int p_97785_, String p_97786_) {
		PoseStack posestack = RenderSystem.getModelViewStack();
		posestack.translate(0.0D, 0.0D, 32.0D);

		RenderSystem.applyModelViewMatrix();
		this.setBlitOffset(200);
		this.itemRenderer.blitOffset = 200.0F;

		net.minecraft.client.gui.Font font = net.minecraftforge.client.RenderProperties.get(p_97783_).getFont(p_97783_);
		if (font == null) font = this.font;

		this.itemRenderer.renderAndDecorateItem(p_97783_, p_97784_, p_97785_);
		this.itemRenderer.renderGuiItemDecorations(font, p_97783_, p_97784_, p_97785_ - (this.draggingItem.isEmpty() ? 0 : 8), p_97786_);

		this.setBlitOffset(0);
		this.itemRenderer.blitOffset = 0.0F;
	}

	@Override
	protected void renderBg(PoseStack p_97787_, float p_97788_, int p_97789_, int p_97790_) {

	}

	@Override
	public void render(PoseStack pose, int mouse_x, int mouse_y, float flag) {
		if (!made_main_frame) {
			made_main_frame = true;
			main_frame = makeMainFrame();

			if (main_frame != null) {
				addPanel(main_frame);

				for (var widget : menu.mattery_widgets) {
					if (widget.auto_parent)
						new MatteryWidgetPanel(this, main_frame, widget);
				}

				for (var slot : menu.main_slots) {
					new SlotPanel<>(this, main_frame, slot);
				}
			}

			movePanels();
		}

		int render_x = this.leftPos;
		int render_y = this.topPos;

		// dark background
		this.renderBackground(pose);

		hoveredSlot = null;

		var hovered = false;

		for (int i = panels.size() - 1; i >= 0; i--) {
			var panel = panels.get(i);

			if (hovered) {
				panel.unsetHover();
			} else if (panel.tickHover(mouse_x, mouse_y)) {
				hovered = true;
			}
		}

		RenderSystem.defaultBlendFunc();
		RenderSystem.enableBlend();
		RenderSystem.enableDepthTest();
		RenderSystem.enableTexture();
		RenderSystem.activeTexture(GL_TEXTURE0);

		float depth = -900f;

		for (var panel : panels) {
			RenderSystem.depthFunc(GL_ALWAYS);

			panel.set3DDepth(depth);
			RenderSystem.setShaderColor(1F, 1F, 1F, 1F);
			depth = Math.max(depth, panel.render(pose, mouse_x, mouse_y, flag));
		}

		RenderSystem.depthFunc(GL_LESS);

		net.minecraftforge.common.MinecraftForge.EVENT_BUS.post(new net.minecraftforge.client.event.GuiContainerEvent.DrawBackground(this, pose, mouse_x, mouse_y));

		RenderSystem.disableDepthTest();

		// Screen.super.render
		for (Widget widget : this.renderables) {
			widget.render(pose, mouse_x, mouse_y, flag);
		}
		// /Screen.super.render

		RenderSystem.disableDepthTest();

		net.minecraftforge.common.MinecraftForge.EVENT_BUS.post(new net.minecraftforge.client.event.GuiContainerEvent.DrawForeground(this, pose, mouse_x, mouse_y));

		var itemstack = this.draggingItem.isEmpty() ? this.menu.getCarried() : this.draggingItem;

		if (!itemstack.isEmpty()) {
			int i2 = this.draggingItem.isEmpty() ? 8 : 16;
			String override_count = null;

			if (this.isQuickCrafting && this.quickCraftSlots.size() > 1) {
				itemstack = itemstack.copy();
				itemstack.setCount(this.quickCraftingRemainder);

				if (itemstack.isEmpty()) {
					override_count = ChatFormatting.YELLOW + "0";
				}
			}

			this.renderFloatingItem(itemstack, mouse_x - render_x - 8, mouse_y - render_y - i2, override_count);
		}

		RenderSystem.disableDepthTest();

		// forge's hard limit tooltip Z position to 500 units
		// so if we have anything drawn above 500 units, lift it up
		if (depth > 500) {
			pose.pushPose();
			pose.translate(0, 0, depth - 500);
		}

		if (this.menu.getCarried().isEmpty()) {
			if (this.hoveredSlot != null && this.hoveredSlot.hasItem()) {
				this.renderTooltip(pose, this.hoveredSlot.getItem(), mouse_x, mouse_y);
			} else {
				for (int i = panels.size() - 1; i >= 0; i--) {
					var panel = panels.get(i);

					if (panel.renderTooltips(pose, mouse_x, mouse_y, flag)) {
						break;
					}
				}
			}
		}

		if (depth > 500) {
			pose.popPose();
		}

		RenderSystem.enableDepthTest();
	}

	// AbstractContainerScreen
	public void recalculateQuickCraftRemaining() {
		ItemStack itemstack = this.menu.getCarried();

		if (!itemstack.isEmpty() && this.isQuickCrafting) {
			if (this.quickCraftingType == 2) {
				this.quickCraftingRemainder = itemstack.getMaxStackSize();
			} else {
				this.quickCraftingRemainder = itemstack.getCount();

				for(Slot slot : this.quickCraftSlots) {
					ItemStack itemstack1 = itemstack.copy();
					ItemStack itemstack2 = slot.getItem();
					int i = itemstack2.isEmpty() ? 0 : itemstack2.getCount();
					AbstractContainerMenu.getQuickCraftSlotCount(this.quickCraftSlots, this.quickCraftingType, itemstack1, i);
					int j = Math.min(itemstack1.getMaxStackSize(), slot.getMaxStackSize(itemstack1));
					if (itemstack1.getCount() > j) {
						itemstack1.setCount(j);
					}

					this.quickCraftingRemainder -= itemstack1.getCount() - i;
				}

			}
		}
	}

	public void onSlotClicked(MatterySlot slot, int mouse_button) {
		InputConstants.Key mouseKey = InputConstants.Type.MOUSE.getOrCreate(mouse_button);
		boolean flag = this.minecraft.options.keyPickItem.isActiveAndMatches(mouseKey);

		long i = Util.getMillis();

		this.doubleclick = this.lastClickSlot == slot && i - this.lastClickTime < 250L && this.lastClickButton == mouse_button;
		this.skipNextRelease = false;

		if (mouse_button != 0 && mouse_button != 1 && !flag) {
			this.checkHotbarMouseClicked(mouse_button);
		} else if (!this.isQuickCrafting) {
			if (this.menu.getCarried().isEmpty()) {
				if (this.minecraft.options.keyPickItem.isActiveAndMatches(mouseKey)) {
					this.slotClicked(slot, slot.index, mouse_button, ClickType.CLONE);
				} else {
					boolean shift_click = (InputConstants.isKeyDown(Minecraft.getInstance().getWindow().getWindow(), 340) || InputConstants.isKeyDown(Minecraft.getInstance().getWindow().getWindow(), 344));
					ClickType clicktype = ClickType.PICKUP;

					if (shift_click) {
						this.lastQuickMoved = slot != null && slot.hasItem() ? slot.getItem().copy() : ItemStack.EMPTY;
						clicktype = ClickType.QUICK_MOVE;
					}

					this.slotClicked(slot, slot.index, mouse_button, clicktype);
				}

				this.skipNextRelease = true;
			} else {
				this.isQuickCrafting = true;
				this.quickCraftingButton = mouse_button;
				this.quickCraftSlots.clear();

				if (mouse_button == 0) {
					this.quickCraftingType = 0;
				} else if (mouse_button == 1) {
					this.quickCraftingType = 1;
				} else if (this.minecraft.options.keyPickItem.isActiveAndMatches(mouseKey)) {
					this.quickCraftingType = 2;
				}
			}
		}

		this.lastClickSlot = slot;
		this.lastClickTime = i;
		this.lastClickButton = mouse_button;
	}

	public void checkHotbarMouseClicked(int mouse_code) {
		if (this.hoveredSlot != null && this.menu.getCarried().isEmpty()) {
			if (this.minecraft.options.keySwapOffhand.matchesMouse(mouse_code)) {
				this.slotClicked(this.hoveredSlot, this.hoveredSlot.index, 40, ClickType.SWAP);
				return;
			}

			for (int i = 0; i < 9; ++i) {
				if (this.minecraft.options.keyHotbarSlots[i].matchesMouse(mouse_code)) {
					this.slotClicked(this.hoveredSlot, this.hoveredSlot.index, i, ClickType.SWAP);
				}
			}
		}
	}

	public void onSlotMouseDragged(@Nonnull MatterySlot slot, int click_type) {
		ItemStack itemstack = this.menu.getCarried();

		if (this.clickedSlot != null && this.minecraft.options.touchscreen) {
			if (click_type == 0 || click_type == 1) {
				if (this.draggingItem.isEmpty()) {
					if (slot != this.clickedSlot && !this.clickedSlot.getItem().isEmpty()) {
						this.draggingItem = this.clickedSlot.getItem().copy();
					}
				} else if (this.draggingItem.getCount() > 1 && AbstractContainerMenu.canItemQuickReplace(slot, this.draggingItem, false)) {
					long i = Util.getMillis();
					if (this.quickdropSlot == slot) {
						if (i - this.quickdropTime > 500L) {
							this.slotClicked(this.clickedSlot, this.clickedSlot.index, 0, ClickType.PICKUP);
							this.slotClicked(slot, slot.index, 1, ClickType.PICKUP);
							this.slotClicked(this.clickedSlot, this.clickedSlot.index, 0, ClickType.PICKUP);
							this.quickdropTime = i + 750L;
							this.draggingItem.shrink(1);
						}
					} else {
						this.quickdropSlot = slot;
						this.quickdropTime = i;
					}
				}
			}
		} else if (this.isQuickCrafting && !itemstack.isEmpty() && (itemstack.getCount() > this.quickCraftSlots.size() || this.quickCraftingType == 2) && AbstractContainerMenu.canItemQuickReplace(slot, itemstack, true) && slot.mayPlace(itemstack) && this.menu.canDragTo(slot)) {
			this.quickCraftSlots.add(slot);
			this.recalculateQuickCraftRemaining();
		}
	}

	public void onMouseReleasedSlot(@Nullable MatterySlot slot, double p_97812_, double p_97813_, int p_97814_, boolean outside) {
		int i = this.leftPos;
		int j = this.topPos;
		InputConstants.Key mouseKey = InputConstants.Type.MOUSE.getOrCreate(p_97814_);
		int k = outside ? -999 : slot != null ? slot.index : -1;

		if (this.doubleclick && p_97814_ == 0 && slot != null && this.menu.canTakeItemForPickAll(ItemStack.EMPTY, slot)) {
			if (hasShiftDown()) {
				if (!this.lastQuickMoved.isEmpty()) {
					for(Slot slot2 : this.menu.slots) {
						if (slot2 != null && slot2.mayPickup(this.minecraft.player) && slot2.hasItem() && slot2.isSameInventory(slot) && AbstractContainerMenu.canItemQuickReplace(slot2, this.lastQuickMoved, true)) {
							this.slotClicked(slot2, slot2.index, p_97814_, ClickType.QUICK_MOVE);
						}
					}
				}
			} else {
				this.slotClicked(slot, k, p_97814_, ClickType.PICKUP_ALL);
			}

			this.doubleclick = false;
			this.lastClickTime = 0L;
		} else {
			if (this.isQuickCrafting && this.quickCraftingButton != p_97814_) {
				this.isQuickCrafting = false;
				this.quickCraftSlots.clear();
				this.skipNextRelease = true;
				return;
			}

			if (this.skipNextRelease) {
				this.skipNextRelease = false;
				return;
			}

			if (this.clickedSlot != null && this.minecraft.options.touchscreen) {
				if (p_97814_ == 0 || p_97814_ == 1) {
					if (this.draggingItem.isEmpty() && slot != this.clickedSlot) {
						this.draggingItem = this.clickedSlot.getItem();
					}

					boolean flag2 = AbstractContainerMenu.canItemQuickReplace(slot, this.draggingItem, false);
					if (k != -1 && !this.draggingItem.isEmpty() && flag2) {
						this.slotClicked(this.clickedSlot, this.clickedSlot.index, p_97814_, ClickType.PICKUP);
						this.slotClicked(slot, k, 0, ClickType.PICKUP);
						if (this.menu.getCarried().isEmpty()) {
							this.snapbackItem = ItemStack.EMPTY;
						} else {
							this.slotClicked(this.clickedSlot, this.clickedSlot.index, p_97814_, ClickType.PICKUP);
							this.snapbackStartX = Mth.floor(p_97812_ - (double)i);
							this.snapbackStartY = Mth.floor(p_97813_ - (double)j);
							this.snapbackEnd = this.clickedSlot;
							this.snapbackItem = this.draggingItem;
							this.snapbackTime = Util.getMillis();
						}
					} else if (!this.draggingItem.isEmpty()) {
						this.snapbackStartX = Mth.floor(p_97812_ - (double)i);
						this.snapbackStartY = Mth.floor(p_97813_ - (double)j);
						this.snapbackEnd = this.clickedSlot;
						this.snapbackItem = this.draggingItem;
						this.snapbackTime = Util.getMillis();
					}

					this.draggingItem = ItemStack.EMPTY;
					this.clickedSlot = null;
				}
			} else if (this.isQuickCrafting && !this.quickCraftSlots.isEmpty()) {
				this.slotClicked((Slot)null, -999, AbstractContainerMenu.getQuickcraftMask(0, this.quickCraftingType), ClickType.QUICK_CRAFT);

				for(Slot slot1 : this.quickCraftSlots) {
					this.slotClicked(slot1, slot1.index, AbstractContainerMenu.getQuickcraftMask(1, this.quickCraftingType), ClickType.QUICK_CRAFT);
				}

				this.slotClicked((Slot)null, -999, AbstractContainerMenu.getQuickcraftMask(2, this.quickCraftingType), ClickType.QUICK_CRAFT);
			} else if (!this.menu.getCarried().isEmpty()) {
				if (this.minecraft.options.keyPickItem.isActiveAndMatches(mouseKey)) {
					this.slotClicked(slot, k, p_97814_, ClickType.CLONE);
				} else {
					boolean flag1 = k != -999 && (InputConstants.isKeyDown(Minecraft.getInstance().getWindow().getWindow(), 340) || InputConstants.isKeyDown(Minecraft.getInstance().getWindow().getWindow(), 344));

					if (flag1) {
						this.lastQuickMoved = slot.hasItem() ? slot.getItem().copy() : ItemStack.EMPTY;
					}

					this.slotClicked(slot, k, p_97814_, flag1 ? ClickType.QUICK_MOVE : ClickType.PICKUP);
				}
			}
		}

		if (this.menu.getCarried().isEmpty()) {
			this.lastClickTime = 0L;
		}

		this.isQuickCrafting = false;
	}

	public void slotClicked(Slot p_97778_, int p_97779_, int p_97780_, ClickType p_97781_) {
		if (p_97778_ != null) {
			p_97779_ = p_97778_.index;
		}

		this.minecraft.gameMode.handleInventoryMouseClick(this.menu.containerId, p_97779_, p_97780_, p_97781_, this.minecraft.player);
	}

	protected boolean checkHotbarKeyPressed(int p_97806_, int p_97807_) {
		if (this.menu.getCarried().isEmpty() && this.hoveredSlot != null) {
			if (this.minecraft.options.keySwapOffhand.isActiveAndMatches(InputConstants.getKey(p_97806_, p_97807_))) {
				this.slotClicked(this.hoveredSlot, this.hoveredSlot.index, 40, ClickType.SWAP);
				return true;
			}

			for (int i = 0; i < 9; ++i) {
				if (this.minecraft.options.keyHotbarSlots[i].isActiveAndMatches(InputConstants.getKey(p_97806_, p_97807_))) {
					this.slotClicked(this.hoveredSlot, this.hoveredSlot.index, i, ClickType.SWAP);
					return true;
				}
			}
		}

		return false;
	}

	@Override
	protected void containerTick() {
		for (var panel : panels) {
			panel.tick();
		}
	}
}
